#pragma once

#include "my_heap.h"
// #define NDEBUG

double TimeElapsed();
int BuildInstance(string);
void FreeMemory();
inline void Undom(int);
inline void Dom(int);
void Add(int);
void Remove(int);
void ConstructDS();
void ConstructByDegree();
void ConstructByScore();
void ResetCandidate();
void UpdateBestSolution();
int CheckSolution();
void UpdateTargetSize();
int ChooseRemoveV(int);
int ChooseAddV();
void AddInit();
void RemoveInit(int);
void init_increase_dominate(int);
void init_decrease_dominate(int);
void init_removeRedundant();
void removeRedundant();
void LocalSearch();

double TimeElapsed() {
  chrono::steady_clock::time_point finish = chrono::steady_clock::now();
  chrono::duration<double> duration = finish - start;
  return duration.count();
}

int BuildInstance(string filename) {
  string tmp;
  string sign;
  int v, e;
  int v1, v2;

  ifstream infile(filename);
  if (!infile) {
    return 1;
  }

  while (infile.peek() != 'p') {
    getline(infile, tmp);
  }

  infile >> sign >> tmp >> v_num >> e_num;

  edge = new Edge[e_num];
  score = new int[v_num + 1];
  // sub_score = new int[v_num + 1];
  time_stamp = new llong[v_num + 1];
  v_edges = new int *[v_num + 1];
  v_adj = new int *[v_num + 1];
  v_degree = new int[v_num + 1];
  v_threshold = new int[v_num + 1];
  v_in_c = new int[v_num + 1];
  candidate = new int[v_num + 1];
  index_in_candidate = new int[v_num + 1];
  best_v_in_c = new int[v_num + 1];
  //conf_change = new int[v_num + 1];
  //conf_count = new int[v_num + 1];
  dominated = new int[v_num + 1];
  best_dominated = new int[v_num + 1];
  undom_stack = new int[v_num + 1];
  index_in_undom_stack = new int[v_num + 1];
  v_fixed = new int[v_num + 1];

  my_heap = new int[v_num + 1];
  pos_in_my_heap = new int[v_num + 1];
  my_heap_count = 0;

  fill_n(v_degree, v_num + 1, 0);
  fill_n(v_threshold, v_num + 1, 0);
  fill_n(v_in_c, v_num + 1, 0);
  fill_n(score, v_num + 1, 0);
  // fill_n(sub_score, v_num + 1, 0);
  //fill_n(conf_change, v_num + 1, 0);
  fill_n(time_stamp, v_num + 1, 0);
  fill_n(dominated, v_num + 1, 0);
  fill_n(best_dominated, v_num + 1, 0);
  fill_n(v_fixed, v_num + 1, 0);
  
  //inference
  must_in = new int[v_num + 1];
  record_in = new int[v_num + 1];
  fill_n(record_in, v_num + 1, 0);


  for (e = 0; e < e_num; e++) {
    infile >> tmp >> v1 >> v2;
    v_degree[v1]++;
    v_degree[v2]++;

    edge[e].v1 = v1;
    edge[e].v2 = v2;
  }
  infile.close();

  v_adj[0] = 0;
  v_edges[0] = 0;
  for (v = 1; v < v_num + 1; v++) {
    v_adj[v] = new int[v_degree[v]];
    v_edges[v] = new int[v_degree[v]];
  }

  int *v_degree_tmp = new int[v_num + 1];
  fill_n(v_degree_tmp, v_num + 1, 0);

  for (e = 0; e < e_num; e++) {
    v1 = edge[e].v1;
    v2 = edge[e].v2;

    v_edges[v1][v_degree_tmp[v1]] = e;
    v_edges[v2][v_degree_tmp[v2]] = e;

    v_adj[v1][v_degree_tmp[v1]] = v2;
    v_adj[v2][v_degree_tmp[v2]] = v1;

    v_degree_tmp[v1]++;
    v_degree_tmp[v2]++;
  }
  delete[] v_degree_tmp;
  
  
  //inference
    for (v = 1; v < v_num + 1; v++) {
    //inference

    if (v_degree[v] == 0)
    {
    	must_in[must_in_end++]=v;
    	record_in[v]=1;
    }
    else if (v_degree[v] == 1)
    {
    	int n =v_adj[v][0];
    	must_in[must_in_end++]=n;
    	record_in[n]=1;
    }
    
    else if (v_degree[v] == 2)
    {
    	int n1 = v_adj[v][0];
    	int n2 = v_adj[v][1];
    	if (v_degree[n1]>2 && v_degree[n2]==2){
    		if (v_adj[n2][0]==n1 || v_adj[n2][1]==n1){
    			must_in[must_in_end++]=n1;
    			record_in[n1]=1;
    		}
    	}
    	else if (v_degree[n2]>2 && v_degree[n1]==2){
    		if (v_adj[n1][0]==n2 || v_adj[n1][1]==n2){
    			must_in[must_in_end++]=n2;
    			record_in[n2]=1;
    		}
    	}
    }
  }

  return 0;
}

void FreeMemory() {
  int v;
  for (v = 0; v < v_num + 1; v++) {
    delete[] v_adj[v];
    delete[] v_edges[v];
  }

  delete[] dominated;
  delete[] undom_stack;
  delete[] index_in_undom_stack;
  //delete[] conf_change;
  //delete[] conf_count;
  delete[] best_v_in_c;
  delete[] index_in_candidate;
  delete[] candidate;
  delete[] v_in_c;
  delete[] v_degree;
  delete[] v_threshold;
  delete[] v_adj;
  delete[] v_edges;
  delete[] time_stamp;
  // delete[] sub_score;
  delete[] score;
  delete[] edge;
  delete[] my_heap;
  delete[] pos_in_my_heap;
}

inline void Undom(int v) {
  index_in_undom_stack[v] = undom_stack_fill_pointer;
  Push(v, undom_stack);
}

inline void Dom(int v) {
  int index, last_undom_vertex;

  last_undom_vertex = Pop(undom_stack);
  index = index_in_undom_stack[v];
  undom_stack[index] = last_undom_vertex;
  index_in_undom_stack[last_undom_vertex] = index;
}

void init_increase_dominate(int v) {
  if (dominated[v] == 0) {
    --score[v];
    int pos = pos_in_my_heap[v];
    my_heap_remove(pos);
    my_heap_insert(v);
    // all make but one break
    for (int i = 0; i < v_degree[v]; ++i) {
      int u = v_adj[v][i];
      --score[u];
      int pos = pos_in_my_heap[u];
      my_heap_remove(pos);
      my_heap_insert(u);
    }
    Dom(v);
  } else if (dominated[v] == 1) {
    // only two vertex need to update break;
    if (v_in_c[v] == 1) {
      ++score[v];
    }
    // TODO: speed up
    for (int i = 0; i < v_degree[v]; ++i) {
      int u = v_adj[v][i];
      if (v_in_c[u]) {
        ++score[u];
      }
    }
  }
  ++dominated[v];
}

void removeRedundant() {
  int v;
  for (int i = 0; i < candidate_size; ++i) {
    v = candidate[i];
    if (score[v] == 0) {
      Remove(v);
      i--;
    }
  }
}

void init_removeRedundant() {
  int v;
  for (int i = 0; i < candidate_size; ++i) {
    v = candidate[i];
    if (score[v] == 0) {
      RemoveInit(v);
      i--;
    }
  }
}
void init_decrease_dominate(int v) {
  if (dominated[v] == 1) {
    // all score of neighbours are make
    ++score[v];
    for (int i = 0; i < v_degree[v]; ++i) {
      int u = v_adj[v][i];
      ++score[u];
    }
    Undom(v);
  } else if (dominated[v] == 2) {
    // only need to update one break
    if (v_in_c[v]) {
      --score[v];
    } else {
      // TODO: speed up this procedure
      for (int i = 0; i < v_degree[v]; ++i) {
        int u = v_adj[v][i];
        if (v_in_c[u]) {
          --score[u];
        }
      }
    }
  }
  --dominated[v];
}

void increase_dominate(int v) {
  if (dominated[v] == 0) {
    --score[v];
    // all make but one break
    for (int i = 0; i < v_degree[v]; ++i) {
      int u = v_adj[v][i];
      --score[u];
    }
    Dom(v);
  } else if (dominated[v] == 1) {
    if (v_in_c[v] == 1) {
      ++score[v];
    } else {
      // TODO: speed up
      for (int i = 0; i < v_degree[v]; ++i) {
        int u = v_adj[v][i];
        if (v_in_c[u]) {
          ++score[u];
          break;
        }
      }
    }
  }
  // sub score
  /*
  if (dominated[v] == 0) {
    if (!v_in_c[v]) {
      sub_score[v]++;
    }
    for (int i = 0; i < v_degree[v]; i++) {
      int u = v_adj[v][i];
      sub_score[u]++;
    }
  } else if (dominated[v] == 1) {
    sub_score[v]--;
    for (int i = 0; i < v_degree[v]; i++) {
      int u = v_adj[v][i];
      sub_score[u]--;
    }
  } else if (dominated[v] == 2) {
    if (v_in_c[v]) {
      sub_score[v]++;
    }
    for (int i = 0; i < v_degree[v]; ++i) {
      int u = v_adj[v][i];
      if (v_in_c[u]) {
        sub_score[u]++;
      }
    }
  }
  */
  ++dominated[v];
  
  /*if (dominated[v] == 1) {
    for (int i = 0; i < v_degree[v]; ++i) {
      int u = v_adj[v][i];
      if (conf_change[u] <= v_threshold[u]) {
        conf_change[u]++;
      }
    }
  }*/
}

bool test_score();
void Add(int v) {
  int new_score = -score[v];
  // int new_subscore = -sub_score[v];

  increase_dominate(v);
  for (int i = 0; i < v_degree[v]; ++i) {
    int u = v_adj[v][i];

    increase_dominate(u);
  }
  score[v] = new_score;
  v_in_c[v] = 1;
  c_size++;
  candidate[candidate_size] = v;
  index_in_candidate[v] = candidate_size++;

  // sub_score[v] = new_subscore;
}

void decrease_dominate(int v) {
  if (dominated[v] == 1) {
    // all score of neighbours are make
    ++score[v];
    for (int i = 0; i < v_degree[v]; ++i) {
      int u = v_adj[v][i];
      ++score[u];
    }
    Undom(v);
  } else if (dominated[v] == 2) {
    // only need to update one break
    if (v_in_c[v]) {
      --score[v];
    } else {
      // TODO: speed up this procedure
      for (int i = 0; i < v_degree[v]; ++i) {
        int u = v_adj[v][i];
        if (v_in_c[u]) {
          --score[u];
          break;
        }
      }
    }
  }
  // sub score
  /*
  if (dominated[v] == 3) {
    if (v_in_c[v]) {
      sub_score[v]--;
    }
    for (int i = 0; i < v_degree[v]; i++) {
      int u = v_adj[v][i];
      if (v_in_c[u] == 1) {
        sub_score[u]--;
      }
    }
  } else if (dominated[v] == 2) {
    sub_score[v]++;
    for (int i = 0; i < v_degree[v]; i++) {
      int u = v_adj[v][i];
      sub_score[u]++;
    }
  } else if (dominated[v] == 1) {
    if (!v_in_c[v]) {
      sub_score[v]--;
    }
    for (int i = 0; i < v_degree[v]; ++i) {
      int u = v_adj[v][i];
      if (!v_in_c[u]) {
        sub_score[u]--;
      }
    }
  }
  */
  --dominated[v];
  /*
  if (dominated[v] == 0) {
    for (int i = 0; i < v_degree[v]; ++i) {
      int u = v_adj[v][i];
      /*if (conf_change[u] <= v_threshold[u]) {
        conf_change[u]++;
      }
    }
  }*/
}

void RemoveInit(int v) {
  v_in_c[v] = 0;
  c_size--;
  int last_candidate_v = candidate[--candidate_size];
  int index = index_in_candidate[v];
  candidate[index] = last_candidate_v;
  index_in_candidate[last_candidate_v] = index;
  score[v] = 0;
  init_decrease_dominate(v);
  for (int i = 0; i < v_degree[v]; ++i) {
    int u = v_adj[v][i];

    init_decrease_dominate(u);
  }
}

void Remove(int v) {
  v_in_c[v] = 0;
  c_size--;
  int last_candidate_v = candidate[--candidate_size];
  int index = index_in_candidate[v];
  candidate[index] = last_candidate_v;
  index_in_candidate[last_candidate_v] = index;

  // from break to 0, then from 0 to break;
  int new_score = -score[v];
  // int new_subscore = -sub_score[v];
  decrease_dominate(v);
  int neighbours = v_degree[v];
  for (int i = 0; i < neighbours; ++i) {
    int u = v_adj[v][i];

    decrease_dominate(u);
  }
  score[v] = new_score;
  // sub_score[v] = new_subscore;
  /*if (step - time_stamp[v] < 10) {
    conf_change[v] = 0;
  }*/
}

bool test_score() {
  vector<int> correct_score(v_num + 1, 0);
  vector<int> correct_dominate(v_num + 1, 0);
  for (int v = 1; v < v_num + 1; ++v) {
    if (v_in_c[v]) {
      correct_dominate[v]++;
      for (int i = 0; i < v_degree[v]; ++i) {
        int u = v_adj[v][i];
        correct_dominate[u]++;
      }
    }
  }
  for (int v = 1; v < v_num + 1; ++v) {
    if (correct_dominate[v] == 0) {
      correct_score[v]++;
      for (int i = 0; i < v_degree[v]; ++i) {
        int u = v_adj[v][i];
        correct_score[u]++;
      }
    } else if (correct_dominate[v] == 1) {
      if (v_in_c[v]) {
        correct_score[v]--;
      } else {
        for (int i = 0; i < v_degree[v]; ++i) {
          int u = v_adj[v][i];
          if (v_in_c[u]) {
            correct_score[u]--;
          }
        }
      }
    }
  }
  bool right = true;
  for (int v = 1; v < v_num + 1; ++v) {
    if (correct_dominate[v] != dominated[v]) {
      std::cout << "dominated wrong: " << v << std::endl;
      std::cout << "correct: " << correct_dominate[v]
                << ", wrong: " << dominated[v] << std::endl;
      right = false;
    }
    if (correct_score[v] != score[v]) {
      std::cout << "score wrong: " << v << std::endl;
      std::cout << "correct: " << correct_score[v] << ", wrong: " << score[v]
                << std::endl;
      std::cout << v_in_c[v] << std::endl;
      right = false;
    }
  }
  if (!right)
    abort();

  right = true;
  int undom_num = 0;
  for (int v = 1; v < v_num + 1; ++v) {
    if (dominated[v] == 0) {
      undom_num++;
      bool okay = false;
      for (int i = 0; i < undom_stack_fill_pointer; ++i) {
        if (undom_stack[i] == v) {
          okay = true;
          break;
        }
      }
      if (!okay) {
        std::cout << "not in undom_stack: " << v << std::endl;
        right = false;
      }
    }
  }
  if (undom_num != undom_stack_fill_pointer) {
    std::cout << "undom_num wrong!" << std::endl;
  }
  if (!right)
    abort();

  return true;
}

void ResetCandidate() {
  int v;
  int j = 0;

  for (v = 1; v < v_num + 1; v++) {
    if (v_in_c[v] == 1 && v_fixed[v] == 0) {
      candidate[j] = v;
      index_in_candidate[v] = j;
      j++;
    } else {
      index_in_candidate[v] = 0;
    }
  }

  candidate_size = j;
}

void UpdateBestSolution() {
  int v;

  if (c_size < best_c_size) {
    for (v = 1; v < v_num + 1; v++) {
      best_v_in_c[v] = v_in_c[v];
      best_dominated[v] = dominated[v];
    }
    best_c_size = c_size;
    best_comp_time = TimeElapsed();
    best_step = step;
  }

}

bool CompareBySecond(const pair<int, int> &lhs, const pair<int, int> &rhs) {
  return lhs.second > rhs.second;
}

void AddInit(int v) {
  int pos;

  v_in_c[v] = 1;

  int new_score = -score[v];
  init_increase_dominate(v);
  for (int i = 0; i < v_degree[v]; ++i) {
    int u = v_adj[v][i];
    init_increase_dominate(u);
  }
  score[v] = new_score;
  pos = pos_in_my_heap[v];
  my_heap_remove(pos);
}

void ConstructByDegree() {
  fill_n(v_in_c, v_num + 1, 0);
  fill_n(score, v_num + 1, 0);
  fill_n(dominated, v_num + 1, 0);
  fill_n(v_fixed, v_num + 1, 0);//cai
  int v;
  // step = 0;
  int ds_size = 0;

  /*for (v = 1; v < v_num + 1; v++) {
    // aovoid duplicate, lin jinkun
    if (dominated[v] > 0) {
      continue;
    }

    if (v_degree[v] == 0) {
      v_in_c[v] = 1;
      dominated[v]++;
      ds_size++;
      time_stamp[v] = (llong)(~0ULL >> 1);
      v_fixed[v] = 1;
    } else if (v_degree[v] == 1) {
      int n = v_adj[v][0];
      v_in_c[n] = 1;
      dominated[n]++;
      ds_size++;
      time_stamp[n] = (llong)(~0ULL >> 1);
      v_fixed[n] = 1;
      for (int i = 0; i < v_degree[n]; i++) {
        dominated[v_adj[n][i]]++;
      }
    }
  }*/
  must_in_beg=0;
  while(must_in_beg<must_in_end)
  {
  	int u = must_in[must_in_beg];
  	must_in_beg++;
  	if (v_in_c[u]==1 || dominated[u] > 0) {
      		continue;
    	}
    	ds_size++;
    	dominated[u]++;
    	
      //AddInit(u);
      v_in_c[u]=1;
      time_stamp[u] = (llong)(~0ULL >> 1);
      v_fixed[u] = 1;
      
      for (int i = 0; i < v_degree[u]; i++) {
        dominated[v_adj[u][i]]++;
      }
  }

  // TODO: speed up by using bin sort
  vector<pair<int, int>> v_d_vec(v_num + 1);
  for (v = 1; v < v_num + 1; v++) {
    v_d_vec[v] = pair<int, int>(v, v_degree[v]);
  }
  sort(v_d_vec.begin() + 1, v_d_vec.end(), CompareBySecond);

  for (v = 1; v < v_num + 1; v++) {
    int v0 = v_d_vec[v].first;
    if (dominated[v0] == 0) {
      dominated[v0]++;
      v_in_c[v0] = 1;
      for (int n = 0; n < v_degree[v0]; n++) {
        dominated[v_adj[v0][n]]++;
      }
      ds_size++;
    }
  }

  // score
  for (v = 1; v < v_num + 1; v++) {
    if (v_in_c[v]) {
      // lin jinkun
      if (dominated[v] == 1) {
        --score[v];
      }
      for (int n = 0; n < v_degree[v]; ++n) {
        if (dominated[v_adj[v][n]] == 1) {
          --score[v];
        }
      }
    }
  }
  c_size = ds_size;
  ResetCandidate();
  init_removeRedundant();
  if (c_size < best_c_size) {
    // std::cout<<"1"<<endl;
    UpdateBestSolution();
  } else // recover related values
  {
    c_size = best_c_size;
    for (v = 1; v < v_num + 1; v++) {
      v_in_c[v] = best_v_in_c[v];
      dominated[v] = best_dominated[v];
      // time_stamp[v] = 0;
    }
    ResetCandidate();
  }
  fill_n(score, v_num + 1, 0);
  // score
  for (v = 1; v < v_num + 1; v++) {
    if (v_in_c[v]) {
      // lin jinkun
      if (dominated[v] == 1) {
        --score[v];
      }
      for (int n = 0; n < v_degree[v]; ++n) {
        if (dominated[v_adj[v][n]] == 1) {
          --score[v];
        }
      }
    }
  }
  // sub_score hwy
  // fill_n(sub_score,v_num + 1,0);
  /*
  for (v = 1; v < v_num + 1; v++) {
    if (v_in_c[v]) {
      if (dominated[v] == 2) {
        --sub_score[v];
      }
      for (int n = 0; n < v_degree[v]; ++n) {
        if (dominated[v_adj[v][n]] == 2) {
          --sub_score[v];
        }
      }

    } else {
      if (dominated[v] == 1) {
        sub_score[v]++;
      }
      for (int n = 0; n < v_degree[v]; ++n) {
        if (dominated[v_adj[v][n]] == 1) {
          sub_score[v]++;
        }
      }
    }
  }
  */
  // UpdateBestSolution();
}

void ConstructByScore() {
  best_c_size = (int)(~0U >> 1);
  int v, i = 0;
  undom_stack_fill_pointer = 0;
  for (v = 1; v < v_num + 1; ++v) {
    score[v] = v_degree[v] + 1;
    v_threshold[v] = v_degree[v] / 20;
    if (v_threshold[v] == 0) {
      v_threshold[v] = 1;
    }

    //conf_change[v] = v_threshold[v] + 1;
    time_stamp[v] = 0;
    Undom(v);
  }
  int *best_array = new int[v_num + 1];
  for (v = 1; v < v_num + 1; ++v) {
    my_heap_insert(v);
  }
  
  
  /*for (v = 1; v < v_num + 1; ++v) {
    // aovoid duplicate, lin jinkun
    if (dominated[v] > 0) {
      continue;
    }
    if (v_degree[v] == 0 && v_in_c[v] == 0) {
      i++;
      AddInit(v);
      time_stamp[v] = (llong)(~0ULL >> 1);
      v_fixed[v] = 1;
    } else if (v_degree[v] == 1) {
      int n = v_adj[v][0];
      if (!v_in_c[n]) {
        AddInit(n);
        i++;
        time_stamp[n] = (llong)(~0ULL >> 1);
        v_fixed[n] = 1;
      }
    }
  }*/
  must_in_beg=0;
  while(must_in_beg<must_in_end)
  {
  	int u = must_in[must_in_beg];
  	must_in_beg++;
  	if (v_in_c[u]==1 || dominated[u] > 0) {
      		continue;
    	}
    	i++;
      AddInit(u);
      time_stamp[u] = (llong)(~0ULL >> 1);
      v_fixed[u] = 1;
  }
  
  
  while (undom_stack_fill_pointer > 0) {
    int best_v = my_heap[0];
    if (score[best_v] > 0) {
      AddInit(best_v);
      i++;
    }
  }
  delete[] best_array;
  c_size = i;
  ResetCandidate();
  init_removeRedundant();

  // ResetCandidate();
  UpdateBestSolution();
}

void ConstructDS() {
  ConstructByScore();
  ConstructByDegree();
}

int CheckSolution() {
  int v;

  for (v = 1; v < v_num + 1; v++) {
    //    	if(best_v_in_c[v])
    //    	{
    //    		check_size++;
    //		}
    if (best_dominated[v] == 0) {
      cout << ", undominated vertex " << v;
      return 0;
    }
  }
  return 1;
}

void UpdateTargetSize() {
  int i, v;
  int best_score;
  int best_remove_v;

  best_remove_v = candidate[0];
  best_score = score[best_remove_v];

  if (score[best_remove_v] != 0) {
    for (i = 1; i < candidate_size; i++) {
      v = candidate[i];

      if (score[v] == 0) {
        // lin jinkun
        best_remove_v = v;
        break;
      }
      if (score[v] > score[best_remove_v]) {
        best_remove_v = v;
      }
    }
  }

  Remove(best_remove_v);
  time_stamp[best_remove_v] = step;
}

class Sub_score {
public:
  int operator[](int v) {
    int ss = 0;
    if (!v_in_c[v]) {
      if (dominated[v] == 1) {
        ss++;
      }
      for (int i = 0; i < v_degree[v]; ++i) {
        int u = v_adj[v][i];
        int d = dominated[u];
        if (d == 1) {
          ss++;
        }
      }
    } else {
      if (dominated[v] == 2) {
        ss--;
      }
      for (int i = 0; i < v_degree[v]; ++i) {
        int u = v_adj[v][i];
        int d = dominated[u];
        if (d == 2) {
          ss--;
        }
      }
    }
    return ss;
  }
};
Sub_score sub_score;

int ChooseRemoveV(int count = 45) {
  int i, v;
  int best_v = candidate[rand() % candidate_size];
  for (i = 0; i < count; i++) {
    v = candidate[rand() % candidate_size];
    if (score[v] > score[best_v]) {
      best_v = v;
    } else if (score[v] == score[best_v] && time_stamp[v] < time_stamp[best_v]) {
    	best_v = v;
    }
  }
  return best_v;
}

int ChooseAddV() {
  // TODO: proof there is at least on avaible add_v
  int base_v, add_v;
  int best_score = -1;
  int best_add_v = -1;

  for (int i = 0; i < undom_stack_fill_pointer; ++i) {
    base_v = undom_stack[i];
    for (int j = 0; j < v_degree[base_v]; ++j) {
      add_v = v_adj[base_v][j];
      /*if (conf_change[add_v] <= v_threshold[add_v]) {
        continue;
      }*/
      if (score[add_v] > best_score) {
        best_add_v = add_v;
        best_score = score[add_v];
      } else if (score[add_v] == best_score) {
        if (time_stamp[add_v] < time_stamp[best_add_v]) {
          best_add_v = add_v;
          best_score = score[add_v];
        }
      }
    }
    //if (conf_change[base_v] > v_threshold[base_v]) {
      if (score[base_v] > best_score) {
        best_add_v = base_v;
        best_score = score[base_v];
      } else if (score[base_v] == best_score) {
        if (time_stamp[base_v] < time_stamp[best_add_v]) {
          best_add_v = base_v;
          best_score = score[base_v];
        }
      }
    //}
  }
  // for some special cases
  /*if (best_add_v == -1) {
    base_v = undom_stack[rand() % undom_stack_fill_pointer];
    best_add_v = base_v;
    best_score = score[base_v];
    is_cal = false;

    for (int i = 0; i < v_degree[base_v]; i++) {
      add_v = v_adj[base_v][i];

      if (score[add_v] > best_score) {
        best_add_v = add_v;
        best_score = score[add_v];
        is_cal = false;
      } else if (score[add_v] == best_score) {
        if (!is_cal) {
          best_sec_score =
              0.1 * sub_score[best_add_v] + (step - time_stamp[best_add_v]);
          is_cal = true;
        }
        sec_score = 0.1 * sub_score[add_v] + (step - time_stamp[add_v]);
        if (sec_score > best_sec_score) {
          best_add_v = add_v;
          best_score = score[add_v];
          best_sec_score = sec_score;
          is_cal = true;
        }
      }
    }
  }*/
  return best_add_v;
}

void LocalSearch() {
  int remove_v, add_v;
  step = 1LL;
  try_step = 100;
  int flag = 0;

  while (true) {
    if (undom_stack_fill_pointer == 0) {
      UpdateBestSolution();
      UpdateTargetSize();
      flag = 0;
      continue;
    }
    if (step % try_step == 0) {
      if (TimeElapsed() >= cutoff_time) {
        return;
      }
    }
    remove_v = ChooseRemoveV();
    Remove(remove_v);
    time_stamp[remove_v] = step;
    if(flag == 0){
        remove_v = candidate[rand() % candidate_size];
        //remove_v = ChooseRemoveV();
    	Remove(remove_v);
    	time_stamp[remove_v] = step;
    	flag = 1;
	}
    if (undom_stack_fill_pointer != 0) {
      add_v = ChooseAddV();
      Add(add_v);
      time_stamp[add_v] = step;
      if(undom_stack_fill_pointer != 0 && flag == 1 && rand() % 100 < para1){
        add_v = ChooseAddV();
      	Add(add_v);
      	time_stamp[add_v] = step;
      	flag = 0;
	  }
    }
    step++;
    if (undom_stack_fill_pointer == 0) {
      removeRedundant();
    }
  }
}
